# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

# Source of truth for Rehydra's version
__version__ = "1.3.4"
from rehydra import utils
from rehydra.errors import MissingConfigException
from rehydra.main import main
from rehydra.types import TaskFunction

from .compose import compose
from .initialize import initialize, initialize_config_dir, initialize_config_module

__all__ = [
    "__version__",
    "MissingConfigException",
    "main",
    "utils",
    "TaskFunction",
    "compose",
    "initialize",
    "initialize_config_module",
    "initialize_config_dir",
]
